# PROTO FBX Export Utils
# 2024 PROTOWLF, Licensed under GPL-3.0
# Some code from Blender v4.2.3 fbx_utils.py, export_fbx_bin.py
#
# Utility functions doing the heavy-lifting of the PROTO FBX export wrapper


import bpy


# From Blender v4.2.3 fbx_utils.py
def ensure_object_not_in_edit_mode(context, obj):
    """Objects in Edit mode usually cannot be exported because much of the API used when exporting is not available for
    Objects in Edit mode.

    Exiting the currently active Object (and any other Objects opened in multi-editing) from Edit mode is simple and
    should be done with `bpy.ops.mesh.mode_set(mode='OBJECT')` instead of using this function.

    This function is for the rare case where an Object is in Edit mode, but the current context mode is not Edit mode.
    This can occur from a state where the current context mode is Edit mode, but then the active Object of the current
    View Layer is changed to a different Object that is not in Edit mode. This changes the current context mode, but
    leaves the other Object(s) in Edit mode.
    """
    if obj.mode != 'EDIT':
        return True

    # Get the active View Layer.
    view_layer = context.view_layer

    # A View Layer belongs to a scene.
    scene = view_layer.id_data

    # Get the current active Object of this View Layer, so we can restore it once done.
    orig_active = view_layer.objects.active

    # Check if obj is in the View Layer. If obj is not in the View Layer, it cannot be set as the active Object.
    # We don't use `obj.name in view_layer.objects` because an Object from a Library could have the same name.
    is_in_view_layer = any(o == obj for o in view_layer.objects)

    do_unlink_from_scene_collection = False
    try:
        if not is_in_view_layer:
            # There might not be any enabled collections in the View Layer, so link obj into the Scene Collection
            # instead, which is always available to all View Layers of that Scene.
            scene.collection.objects.link(obj)
            do_unlink_from_scene_collection = True
        view_layer.objects.active = obj

        # Now we're finally ready to attempt to change obj's mode.
        if bpy.ops.object.mode_set.poll():
            bpy.ops.object.mode_set(mode='OBJECT')
        if obj.mode == 'EDIT':
            # The Object could not be set out of EDIT mode and therefore cannot be exported.
            return False
    finally:
        # Always restore the original active Object and unlink obj from the Scene Collection if it had to be linked.
        view_layer.objects.active = orig_active
        if do_unlink_from_scene_collection:
            scene.collection.objects.unlink(obj)

    return True


# Adapted from Blender v4.2.3 export_fbx_bin.py save()
def get_export_objects(context, use_selection=False, use_visible=False, use_active_collection=False, collection=""):
    """
    Get the objects the user wishes to export
    """
    
    source_collection = None
    if use_active_collection:
        source_collection = context.view_layer.active_layer_collection.collection
    elif collection:
        local_collection = bpy.data.collections.get((collection, None))
        if local_collection:
            source_collection = local_collection
        else:
            operator.report({'ERROR'}, "Collection '%s' was not found" % collection)
            empty = []
            return empty

    if source_collection:
        if use_selection:
            ctx_objects = tuple(obj for obj in source_collection.all_objects if obj.select_get())
        else:
            ctx_objects = source_collection.all_objects
    else:
        if use_selection:
            ctx_objects = context.selected_objects
        else:
            ctx_objects = context.view_layer.objects
    if use_visible:
        ctx_objects = tuple(obj for obj in ctx_objects if obj.visible_get())

    # Ensure no Objects are in Edit mode.
    # Copy to a tuple for safety, to avoid the risk of modifying ctx_objects while iterating.
    for obj in tuple(ctx_objects):
        if not ensure_object_not_in_edit_mode(context, obj):
            operator.report({'ERROR'}, "%s could not be set out of Edit Mode, so cannot be exported" % obj.name)
            empty = []
            return empty

    return ctx_objects


def get_actions_for_armature(armature):
    # Detect all actions that manipulate this armature
    # We have to peek into the fcurve data of each action, to see if we find
    # a bone in our armature
    # NOTE: This is not a perfect solution -- in the future maybe have the user specify which actions to export
    actions = []
    for action in bpy.data.actions:
        valid_action = False
        for fcurve in action.fcurves:
            if fcurve.data_path.startswith("pose.bones"):
                bone_name = fcurve.group.name
                if bone_name in armature.pose.bones:
                    valid_action = True
                break
        if valid_action:
            actions.append(action)
    
    return actions


def action_exports_shapekeys(action):
    if "Export Shapekey Animation" in action and action["Export Shapekey Animation"] == True:
        return True
    return False


def select_objects_for_action_export(context, armature, action):
    # Is this action flagged to export shape keys?
    if action_exports_shapekeys(action):
        #print("Finding shapekey objects")
        # Get all deformed meshes with shape key information
        shapekey_objects = []
        for ob in context.view_layer.objects:
            if ob.type == 'MESH' and ob.data.shape_keys != None:
                #print("Object " + ob.name + " has shapekeys")
                for modifier in ob.modifiers:
                    if modifier.type == 'ARMATURE' and modifier.object.name == armature.name:
                        #print("Object " + ob.name + " added to export")
                        shapekey_objects.append(ob)
        # Select them and the armature
        bpy.ops.object.select_all(action='DESELECT')
        for ob in shapekey_objects:
            ob.select_set(state=True)
        bpy.context.view_layer.objects.active = armature
        armature.select_set(state=True)
    
    else:
        # Only select the armature
        bpy.ops.object.select_all(action='DESELECT')
        bpy.context.view_layer.objects.active = armature
        armature.select_set(state=True)


def duplicate_export_objects(objects, armature_export_name):
    """
    Duplicate objects the user wishes to export, and associate duplicate meshes with any appropriate duplicate armatures
    """
    
    duplicate_objects = []
    mesh_objects = []
    original_to_duplicate_armatures = {}
    armature_original_name = ""
    
    
    for ob in objects:
        # duplicate object
        duplicate_object = ob.copy()
        duplicate_object.data = ob.data.copy()
        
        # link to duplicate to root scene collection (we know it won't be a linked collection!)
        bpy.context.scene.collection.objects.link(duplicate_object)
        bpy.context.view_layer.update()
        
        # Cache mapping between original/duplicate armatures
        if ob.type == 'ARMATURE':
            original_to_duplicate_armatures[ob] = duplicate_object
        
        duplicate_objects.append(duplicate_object)
        
        # For mesh objects, if linked, make the object and its data local
        # NOTE: I was actually able to modify drivers of a library-override object without making them
        # local, which is weird... keeping this here anyway just in-case
        if ob.type == 'MESH':
            mesh_objects.append(duplicate_object)
            if duplicate_object.library != None or duplicate_object.override_library != None:
                bpy.ops.object.select_all(action='DESELECT')
                bpy.context.view_layer.objects.active = duplicate_object
                duplicate_object.select_set(state=True)
                bpy.ops.object.make_local(type='SELECT_OBDATA')
    
    # Rename armatures
    # Original armature gets a temp name (will be renamed later)
    # (This dictionary will only ever be 0 or 1 long, but might as well lay groundwork for eventually supporting multiple armatures)
    for original_armature, duplicate_armature in original_to_duplicate_armatures.items():
        # If option is empty, export with original name
        if armature_export_name == "":
            armature_export_name = original_armature.name
        armature_original_name = original_armature.name
        original_armature.name = armature_export_name + "_temp"
        duplicate_armature.name = armature_export_name
    
    # Associate duplicate meshes with duplicate armatures
    for mesh_ob in mesh_objects:
        # re-parent
        if mesh_ob.parent in original_to_duplicate_armatures:
            #print( "re-parenting dupe object " + mesh_ob.name + " from " + mesh_ob.parent.name + " to " + original_to_duplicate_armatures[mesh_ob.parent].name)
            mesh_ob.parent = original_to_duplicate_armatures[mesh_ob.parent]
            
        # re-associate armature modifier
        for mod in mesh_ob.modifiers:
            if mod.type != 'ARMATURE':
                continue
            if mod.object == None:
                continue
            if mod.object in original_to_duplicate_armatures:
                #print( "re-associating armature modifier for " + mesh_ob.name + " from " + mod.object.name + " to " + original_to_duplicate_armatures[mod.object].name)
                mod.object = original_to_duplicate_armatures[mod.object]
        
        # re-map any shapekey drivers driven by the armature
        shape_keys = mesh_ob.data.shape_keys
        if shape_keys != None and shape_keys.animation_data != None and shape_keys.animation_data.drivers != None:
            drivers = shape_keys.animation_data.drivers
            for driver_fcurve in drivers:
                for var in driver_fcurve.driver.variables:
                    for target in var.targets:
                        for original_armature, duplicate_armature in original_to_duplicate_armatures.items():
                            if target.id == original_armature.id_data:
                                target.id = duplicate_armature.id_data
            
        
    # Update dependencies
    #depsgraph.update()
    bpy.context.evaluated_depsgraph_get().update()
        
    return duplicate_objects, armature_original_name
    

def force_scale_object(ob, unit_scale):
    # unlock scale, in case it's locked
    ob.lock_scale[0] = False
    ob.lock_scale[1] = False
    ob.lock_scale[2] = False
    
    # disable drivers on transform
    disabled_drivers = []
    if ob.animation_data:
        for driver in ob.animation_data.drivers:
            if driver.data_path in ['location', 'rotation_euler', 'rotation_quaternion', 'scale']:
                if driver.mute == False:
                    driver.mute = True
                    disabled_drivers.append(driver)
                    
    # disable constraints
    disabled_constraints = []
    for constraint in ob.constraints:
        if constraint.enabled:
            constraint.enabled = False
            disabled_constraints.append(constraint)
    
    # scale the object
    ob.location *= unit_scale
    ob.scale *= unit_scale
    
    # apply scale
    bpy.ops.object.select_all(action='DESELECT')
    bpy.context.view_layer.objects.active = ob
    ob.select_set(state=True)
    bpy.ops.object.transform_apply(location=False, rotation=False, scale=True)
    
    bpy.context.evaluated_depsgraph_get().update()
    #return disabled_drivers, disabled_constraints
    
    # re-enable drivers on transform
    for driver in disabled_drivers:
        driver.mute = False
    
    # re-enable constraints
    for constraint in disabled_constraints:
        constraint.enabled = True
    

def unit_scale_export_objects(unit_scale, objects, scale_changed_actions):
    """
    Scale export objects, and adjust scene unit scale, to bake unit conversion into the exported objects
    """
    # NOTE: we assume 1 armature, and all objects here are duplicates (we don't need to restore scale changes)
    # Changes to actions do need to be restored
    mesh_objects = []
    armature = None
    for ob in tuple(objects):
        if ob.type == 'ARMATURE':
            armature = ob
        elif ob.type == 'MESH':
            mesh_objects.append(ob)
    
    
    armature_meshes = []
    non_armature_meshes = []
    original_object_parents = {}
    for mesh in mesh_objects:
        # unlock scale, in case it's locked
        mesh.lock_scale[0] = False
        mesh.lock_scale[1] = False
        mesh.lock_scale[2] = False
        
        # Make sure all meshes manipulated by the armature are parented to it
        is_armature_mesh = False
        for modifier in mesh.modifiers:
            if armature != None and modifier.type == 'ARMATURE' and modifier.object.name == armature.name:
                #print("found mesh " + mesh.name + " associated with armature " + armature.name)
                armature_meshes.append(mesh)
                is_armature_mesh = True
                if mesh.parent != armature:
                    original_object_parents[mesh] = mesh.parent
                    mesh.parent = armature
                break
        if is_armature_mesh == False:
            # The mesh did not have an armature modifier using the armature, or we aren't exporting an armature
            non_armature_meshes.append(mesh)
    
    #print("non_armature_meshes:")
    #for ob in non_armature_meshes:
    #    print(ob.name)
    #print("armature_meshes:")
    #for ob in armature_meshes:
    #    print(ob.name)
    
    # scale armature
    if armature != None:
        force_scale_object(armature, unit_scale)
    
    # apply scale on meshes influenced by armature
    #for mesh in armature_meshes:
    #    mesh.transform_apply(location=False, rotation=False, scale=True)
    
    # scale any meshes that are not associated with the armature
    for mesh in non_armature_meshes:
        force_scale_object(mesh, unit_scale)
        
    # scale all armature actions
    for action in bpy.data.actions:
        valid_action = False
        for fcurve in action.fcurves:
            if fcurve.data_path.startswith("pose.bones"):
                valid_action = True
                break
        if valid_action:
            for fcurve in action.fcurves:
                if 'location' in fcurve.data_path:
                    for point in fcurve.keyframe_points:
                        point.co[1] *= unit_scale
                        point.handle_left[1] *= unit_scale
                        point.handle_right[1] *= unit_scale
            scale_changed_actions.append(action)

    # Reset constraints that don't play nicely with scale
    if armature != None:
        for pb in armature.pose.bones:
            for constraint in pb.constraints:
                #if constraint.name == "Stretch To":
                if constraint.type == "STRETCH_TO":
                    # Stretch To
                    #print("reset stretch to")
                    constraint.rest_length = 0
                #elif constraint.name == "Child Of":
                elif constraint.type == "CHILD_OF":
                    # Child Of
                    #print("reset child of")
                    constraint.set_inverse_pending = True
                    #context_py = bpy.context.copy()
                    #context_py["constraint"] = constraint
                    #armature.data.bones.active = pb.bone
                    #bpy.ops.constraint.childof_set_inverse(context_py, constraint="Child Of", owner='BONE')
                
    # Scale mesh modifiers that are influenced by scale
    for mesh in mesh_objects:
        for modifier in mesh.modifiers:
            if modifier.type == "MIRROR":
                modifier.merge_threshold *= unit_scale
            elif modifier.type == "SOLIDIFY":
                modifier.thickness *= unit_scale
            elif modifier.type == "BEVEL":
                modifier.width *= unit_scale
            elif modifier.type == 'SHRINKWRAP':
                if modifier.target:
                    tar_average_scale = (modifier.target.scale[0] + modifier.target.scale[1] + modifier.target.scale[2]) / 3
                    modifier.offset *= tar_average_scale

    # Counter-scale the scene's unit scale
    #bpy.context.scene.unit_settings.system = 'METRIC'
    bpy.context.scene.unit_settings.scale_length = 1 / unit_scale
    
    #bpy.context.evaluated_depsgraph_get().update()
    bpy.context.view_layer.update()
    #depsgraph.update()


def apply_min_two_frames_to_actions(armature, original_action_frame_ranges):
    actions = get_actions_for_armature(armature)
    
    for action in actions:
        frame_start, frame_end = action.frame_range
        if frame_start == frame_end:
            # Cache current settings to restore later
            original_action_frame_ranges[action] = (action.use_frame_range, action.frame_range)
            
            # Set manual frame range that's 1 longer
            action.use_frame_range = True
            action.frame_range = frame_start, frame_end + 1
            
    

def do_fbx_export(filepath, keywords):
    """
    Perform an FBX export using Blender's included FBX exporter, with selected objects
    """
    
    bpy.ops.export_scene.fbx(
        filepath = filepath,
        check_existing = False,
    
        # Include
        use_selection = True,
        use_visible = True,
        use_active_collection = False,
        collection = "",
        batch_mode = 'OFF',
        use_batch_own_dir = False,
        object_types = keywords["object_types"],
    
        # Include
        use_custom_props = keywords["use_custom_props"],
    
        # Transform
        global_scale = keywords["global_scale"],
        apply_scale_options = keywords["apply_scale_options"],
        axis_forward = keywords["axis_forward"],
        axis_up = keywords["axis_up"],
        apply_unit_scale = keywords["apply_unit_scale"],
        bake_space_transform = keywords["bake_space_transform"],
    
        # Geometry
        mesh_smooth_type = keywords["mesh_smooth_type"],
        use_subsurf = keywords["use_subsurf"],
        use_mesh_modifiers  = keywords["use_mesh_modifiers"],
        use_mesh_edges = keywords["use_mesh_edges"],
        use_tspace  = keywords["use_tspace"],
    
        # Armature
        primary_bone_axis = keywords["primary_bone_axis"],
        secondary_bone_axis  = keywords["secondary_bone_axis"],
        armature_nodetype = keywords["armature_nodetype"],
        use_armature_deform_only = keywords["use_armature_deform_only"],
        add_leaf_bones = keywords["add_leaf_bones"],
    
        # Bake Animations
        bake_anim = keywords["bake_anim"],
        bake_anim_use_all_bones = keywords["bake_anim_use_all_bones"],
        bake_anim_use_nla_strips = keywords["bake_anim_use_nla_strips"],
        bake_anim_use_all_actions = keywords["bake_anim_use_all_actions"],
        bake_anim_force_startend_keying = keywords["bake_anim_force_startend_keying"],
        bake_anim_step = keywords["bake_anim_step"],
        bake_anim_simplify_factor = keywords["bake_anim_simplify_factor"],
    )
    
    
def cleanup_scene(context, unit_scale, saved_unit_settings, original_objects, duplicate_objects, armature_original_name, scale_changed_actions, original_action_frame_ranges):
    """
    Delete duplicate objects, revert unit scale changes to scene and actions, undo name changes
    """
    
    # Delete duplicate objects
    for ob in duplicate_objects:
        bpy.data.objects.remove(ob, do_unlink=True)
        
    # revert animation curves location scale (x100 Units)
    if unit_scale != 1.0:
        for action in scale_changed_actions:
            for fcurve in action.fcurves:
                if 'location' in fcurve.data_path:
                    for point in fcurve.keyframe_points:
                        point.co[1] *= 1.0 / unit_scale
                        point.handle_left[1] *= 1.0 / unit_scale
                        point.handle_right[1] *= 1.0 / unit_scale
    
    # revert adjusted action frame ranges
    for action in original_action_frame_ranges:
        original_use_frame_range, original_frame_range = original_action_frame_ranges[action]
        action.use_frame_range = original_use_frame_range
        if action.use_frame_range:
            action.frame_range = original_frame_range
    
    # Undo armature name change
    for ob in original_objects:
        if ob.type == 'ARMATURE':
            ob.name = armature_original_name
    
    # Reset scene scale
    # start by restoring the unit system value, since changing it will unexpectedly reset others to default values
    context.scene.unit_settings.system = saved_unit_settings['system']
    for prop_name in saved_unit_settings:
        if prop_name != 'system':
            value = saved_unit_settings[prop_name]
            try:# if Unit System is None, the attributes may be locked
                setattr(context.scene.unit_settings, prop_name, value)
            except:
                pass
    
